import os

import time
from box import Box

import lgblkb_tools.folder_utils
from lgblkb_tools import TheLogger
from lgblkb_tools.remote_opers import SshFolder
from . import global_support as gsup
from .global_support import run_command

simple_logger=TheLogger(__name__)
flower_port=5553
app_name='celery_worker'

def deploy(sshmanager,lgblkb_tools_version,commit_message,detached=True,downup=True):
	run_command(f'git commit -a -m "{commit_message}"')
	run_command('git push')
	time.sleep(5)
	with sshmanager.cd_to_working_dir():
		# sshmanager.put('configs.yaml','configs.yaml')
		sshmanager.run(f"""
		git stash
		git fetch
		git pull
		sed -i "s/ENV LGBLKB_TOOLS_VERSION=''/ENV LGBLKB_TOOLS_VERSION='{lgblkb_tools_version}'/g" Dockerfile
		docker build -t hub.egistic.kz/imgback:latest .
		
		""")  #--build-arg UID=$(id -u) --build-arg GID=$(id -g)
		if downup:
			sshmanager.run(f"""
		docker-compose down
		docker-compose up {'-d' if detached else ''}""")

def get_updated_lgblkb_tools(dry=False,with_commit=False):
	import git
	version_getter=lambda:Box.from_yaml(filename='/home/lgblkb/PycharmProjects/lgblkb_tools/package_info.yaml').version
	if dry:
		lgblkb_tools_version=version_getter()
	else:
		package_name="lgblkb-tools"
		run_command(f"""
	. ~/.virtualenvs/lgblkb_tools/bin/activate
	cd /home/lgblkb/PycharmProjects/lgblkb_tools
	python ./setup.py sdist bdist_wheel
	twine upload dist/*
			""")
		lgblkb_tools_version=version_getter()
		run_command(f"""
	. ~/.virtualenvs/korsum/bin/activate
	pip install --upgrade {package_name}=={lgblkb_tools_version}
	pip install --upgrade {package_name}=={lgblkb_tools_version}
		""")
		if with_commit:
			repo=git.Repo(lgblkb_tools.folder_utils.Folder().path)
			simple_logger.info('repo: %s',repo)
			assert not repo.bare
			repo.index.commit(f"""Updated lgblkb_tools to version='{lgblkb_tools_version}'""")
	
	simple_logger.info('lgblkb_tools_version: %s',lgblkb_tools_version)
	return lgblkb_tools_version

def get_current_branch_name(repo):
	branches=list(map(lambda x:x.strip(),repo.git.branch().split('\n')))
	simple_logger.debug('branches: %s',branches)
	
	for branch in branches:
		if '* ' in branch:
			the_branch=branch[2:]
			simple_logger.info('Current branch: %s',the_branch)
			return the_branch

def get_corresponding_path(base_dir,path):
	path_parts=gsup.get_splitted(path)[1:]
	for i,path_part in enumerate(path_parts):
		if os.path.exists(os.path.join(base_dir,path_part)):
			return os.path.join(base_dir,*path_parts[i:])

class RemoteProjectFolder(SshFolder):
	
	def __init__(self,remote_path,local_project_folder,**conn_kwargs):
		super(RemoteProjectFolder,self).__init__(remote_path=remote_path,**conn_kwargs)
		self.local_project_folder=local_project_folder
		pass
	
	def _download_remote_item(self,remote_itempath,local_itempath=''):
		remote_log_path=get_corresponding_path(self.path,remote_itempath)
		local_itempath=local_itempath or get_corresponding_path(self.local_project_folder.path,remote_itempath)
		os.makedirs(gsup.get_parent_dir(local_itempath),exist_ok=True)
		self.get(remote_log_path,local_itempath)
		return local_itempath
	
	def run_in_docker(self,cmd):
		output=self.run_with_cd(rf'docker run -i --rm --name lgblkb_test_2 '
		                        rf'--network=redis_net '
		                        rf'--network=postgis '
		                        rf'-v "$(pwd)":/app '
		                        rf'-v /home/egistic_db/egistic/image_backend:/image_backend '
		                        rf'-v /nfs/storage/sat_images:/app/nfs_storage_images '
		                        rf'hub.egistic.kz/imgback:latest {cmd}',
		                        warn=True)
		return output
	
	def upload_and_run_in_docker(self,module,args_cmd: str):
		module_path=module.__name__.replace('.','/')+'.py'
		self.put(module.__file__,module_path)
		self.put('configs.yaml','configs.yaml')
		return self.run_in_docker(rf"""python3 -m {module.__name__} {args_cmd}""")
	
	def _upload_and_run(self,python_module,cmd):
		helper_script=python_module.__name__.replace('.','/')+'.py'
		self.put(python_module.__file__,helper_script)
		output=self.run_with_cd(f"""
		docker run -i --rm --name lgblkb_test -v "$(pwd)":/app hub.egistic.kz/imgback:latest {cmd}
		""")
		self.run_with_cd(f'rm {helper_script}',warn=True)
		return output
	
	def commit_and_deploy(self,image_tag,update_lgblkb_tools=False,downup=True,detached=True,):
		MESSAGE=input('Commit message: ')
		lgblkb_tools_version=get_updated_lgblkb_tools(dry=not update_lgblkb_tools)
		# deploy(self.sshman,lgblkb_tools_version,MESSAGE,detached=detached,downup=downup)
		run_command(f'git commit -a -m "{MESSAGE}"')
		run_command('git push')
		# time.sleep(5)
		self.put('configs.yaml','configs.yaml')
		self.put('.env','.env')
		with self.sshman.cd_to_working_dir():
			self.sshman.run(f"""
				git stash
				git fetch
				git pull
				sed -i "s/ENV LGBLKB_TOOLS_VERSION=''/ENV LGBLKB_TOOLS_VERSION='{lgblkb_tools_version}'/g" Dockerfile
				docker build -t {image_tag} .
				""")  #--build-arg UID=$(id -u) --build-arg GID=$(id -g)
			
			if downup:
				self.sshman.run(f"""
				docker-compose down
				docker-compose up {'-d' if detached else ''}""")
	
	def find_log_with_content(self,content):
		self.run_with_cd(r"""

		""")

def open_in_pycharm(filepath):
	run_command(f'/snap/bin/pycharm-professional "{filepath}"')

def main():
	pass

if __name__=='__main__':
	main()



import logging.handlers as log_handlers
from timeit import default_timer as timer
import functools
from box import Box
from colorlog import ColoredFormatter
import logging
from python_log_indenter import IndentedLoggerAdapter

colored_log_format=f"%(log_color)s%(asctime)s %(processName)18.18s %(levelname)5.5s || %(message)s"
base_log_indent_settings=dict(indent_char='| ',spaces=1)
date_fmt="%m-%d %H:%M:%S"
wrapper_log_level=logging.INFO
default_log_level=logging.INFO
default_log_format=colored_log_format.replace(r'%(log_color)s','')
log_func_mapper=Box()
log_func_mapper[logging.DEBUG]=lambda logger:logger.debug
log_func_mapper[logging.INFO]=lambda logger:logger.info
log_func_mapper[logging.WARNING]=lambda logger:logger.warning
log_func_mapper[logging.CRITICAL]=lambda logger:logger.critical
log_func_mapper[logging.ERROR]=lambda logger:logger.error

log_sayer=lambda logger,log_level:log_func_mapper[log_level](logger)

class TheLogger(IndentedLoggerAdapter):
	
	def __init__(self,name,extra=None,auto_add=True,log_level=None,log_format='',**kwargs):
		logger=setup_logger(name,log_level,log_format)
		logger.propagate=False
		super(TheLogger,self).__init__(logger,extra,auto_add,**dict(base_log_indent_settings,**kwargs))
		self.current_handlers=[]
	
	def add_file_handler(self,log_filepath,log_level=logging.DEBUG):
		log_handler=logging.FileHandler(log_filepath)
		log_handler.setFormatter(logging.Formatter(default_log_format))
		log_handler.setLevel(log_level)
		self.logger.addHandler(log_handler)
		self.current_handlers.append(log_handler)
		return self
	
	def add_rotating_handler(self,log_filepath,log_level=None,maxBytes=1.99e6,backupCount=14,**other_opts):
		log_level=log_level or default_log_level
		log_handler=log_handlers.RotatingFileHandler(maxBytes=int(maxBytes),filename=log_filepath,backupCount=backupCount,**other_opts)
		log_handler=self.logger.addHandler(log_handler,level=log_level,log_format=default_log_format)
		self.current_handlers.append(log_handler)
		return self
	
	def wrap(self,log_level=None,skimpy=False):
		def second_wrapper(func):
			@functools.wraps(func)
			def wrapper(*args,**kwargs):
				log_say=log_sayer(self,log_level or wrapper_log_level)
				if skimpy: log_say('Performing "%s"...',func.__name__)
				else: log_say('Running "%s":',func.__name__)
				start=timer()
				self.add()
				result=func(*args,**kwargs)
				if not skimpy: log_say('Done "%s". Duration: %.3f sec.',func.__name__,timer()-start)
				self.sub()
				return result
			
			return wrapper
		
		return second_wrapper
	
	def remove_active_handler(self):
		target_handler=self.current_handlers.pop()
		self.logger.handlers.remove(target_handler)
		return self

def setup_logger(name,log_level=None,log_format=''):
	log_level=log_level or default_log_level
	new_logger=logging.getLogger(name)
	new_logger.setLevel(log_level)
	log_handler=logging.StreamHandler()
	log_format=log_format or colored_log_format
	log_handler.setFormatter(ColoredFormatter(log_format,date_fmt))
	new_logger.addHandler(log_handler)
	# new_logger.info(f'Root logger initialized with level {logging._levelToName[log_level]}.')
	return new_logger

# root_logger=setup_logger(log_level=logging.DEBUG)

simple_logger=TheLogger(__name__)
# log_wrapper=lambda log_level=logging.DEBUG,skimpy=False:simple_logger.wrap(log_level=log_level,skimpy=skimpy)
log_wrapper=simple_logger.wrap

# @simple_logger.wrap(log_level=logging.DEBUG)
@log_wrapper()
def main():
	x=[1,2,3]
	simple_logger.info('x: %s',x)
	
	
	pass

if __name__=='__main__':
	main()



import circus
import signal
from circus.watcher import Watcher

# watcher_conf_override=dict(numprocesses=None,
#                            warmup_delay=None,
#                            working_dir=None,
#                            shell=None,
#                            shell_args=None,
#                            uid=None,
#                            max_retry=None,
#                            gid=None,
#                            send_hup=None,
#                            stop_signal=None,
#                            stop_children=None,
#                            env=None,
#                            graceful_timeout=None,
#                            prereload_fn=None,
#                            rlimits=None,
#                            executable=None,
#                            stdout_stream=None,
#                            stderr_stream=None,
#                            priority=None,
#                            loop=None,
#                            singleton=None,
#                            use_sockets=None,
#                            copy_env=None,
#                            copy_path=None,
#                            max_age=None,
#                            max_age_variance=None,
#                            hooks=None,
#                            respawn=None,
#                            autostart=None,
#                            on_demand=None,
#                            virtualenv=None,
#                            stdin_socket=None,
#                            close_child_stdin=None,
#                            close_child_stdout=None,
#                            close_child_stderr=None,
#                            virtualenv_py_ver=None,
#                            use_papa=None,
#                            )

watcher_conf_override={'numprocesses':None,'warmup_delay':None,'working_dir':None,'shell':None,'shell_args':None,'uid':None,'max_retry':None,'gid':None,'send_hup':None,
                       'stop_signal':None,'stop_children':None,'env':None,'graceful_timeout':None,'prereload_fn':None,'rlimits':None,'executable':None,'stdout_stream':None,
                       'stderr_stream':None,'priority':None,'loop':None,'singleton':None,'use_sockets':None,'copy_env':None,'copy_path':None,'max_age':None,'max_age_variance':None,
                       'hooks':None,'respawn':None,'autostart':None,'on_demand':None,'virtualenv':None,'stdin_socket':None,'close_child_stdin':None,'close_child_stdout':None,
                       'close_child_stderr':None,'virtualenv_py_ver':None,'use_papa':None}

class TheWatcher(object):
	watchers=list()
	
	def __init__(self,name,cmd,numprocesses=1,warmup_delay=0.,working_dir=None,shell=True,shell_args=None,uid=None,max_retry=5,gid=None,
	             send_hup=False,
	             stop_signal=signal.SIGTERM,stop_children=False,env=None,graceful_timeout=30.0,prereload_fn=None,rlimits=None,executable=None,
	             stdout_stream=None,stderr_stream=None,priority=0,loop=None,singleton=False,use_sockets=False,copy_env=False,copy_path=False,
	             max_age=0,max_age_variance=30,hooks=None,respawn=False,autostart=True,on_demand=False,virtualenv=None,stdin_socket=None,
	             close_child_stdin=True,close_child_stdout=False,close_child_stderr=False,virtualenv_py_ver=None,use_papa=False):
		# working_dir=working_dir or default_working_dir
		# virtualenv=virtualenv or virtualenv_root
		# warmup_delay=warmup_delay or watcher_warmup_delay
		# if watcher_respawn is False:
		# 	respawn=watcher_respawn
		if virtualenv: copy_env=True
		print(cmd)
		self.cmd=cmd
		self.name=name
		# region Fields:
		self.numprocesses=numprocesses if watcher_conf_override['numprocesses'] is None else watcher_conf_override['numprocesses']
		self.warmup_delay=warmup_delay if watcher_conf_override['warmup_delay'] is None else watcher_conf_override['warmup_delay']
		self.working_dir=working_dir if watcher_conf_override['working_dir'] is None else watcher_conf_override['working_dir']
		self.shell=shell if watcher_conf_override['shell'] is None else watcher_conf_override['shell']
		self.shell_args=shell_args if watcher_conf_override['shell_args'] is None else watcher_conf_override['shell_args']
		self.uid=uid if watcher_conf_override['uid'] is None else watcher_conf_override['uid']
		self.max_retry=max_retry if watcher_conf_override['max_retry'] is None else watcher_conf_override['max_retry']
		self.gid=gid if watcher_conf_override['gid'] is None else watcher_conf_override['gid']
		self.send_hup=send_hup if watcher_conf_override['send_hup'] is None else watcher_conf_override['send_hup']
		self.stop_signal=stop_signal if watcher_conf_override['stop_signal'] is None else watcher_conf_override['stop_signal']
		self.stop_children=stop_children if watcher_conf_override['stop_children'] is None else watcher_conf_override['stop_children']
		self.env=env if watcher_conf_override['env'] is None else watcher_conf_override['env']
		self.graceful_timeout=graceful_timeout if watcher_conf_override['graceful_timeout'] is None else watcher_conf_override['graceful_timeout']
		self.prereload_fn=prereload_fn if watcher_conf_override['prereload_fn'] is None else watcher_conf_override['prereload_fn']
		self.rlimits=rlimits if watcher_conf_override['rlimits'] is None else watcher_conf_override['rlimits']
		self.executable=executable if watcher_conf_override['executable'] is None else watcher_conf_override['executable']
		self.stdout_stream=stdout_stream if watcher_conf_override['stdout_stream'] is None else watcher_conf_override['stdout_stream']
		self.stderr_stream=stderr_stream if watcher_conf_override['stderr_stream'] is None else watcher_conf_override['stderr_stream']
		self.priority=priority if watcher_conf_override['priority'] is None else watcher_conf_override['priority']
		self.loop=loop if watcher_conf_override['loop'] is None else watcher_conf_override['loop']
		self.singleton=singleton if watcher_conf_override['singleton'] is None else watcher_conf_override['singleton']
		self.use_sockets=use_sockets if watcher_conf_override['use_sockets'] is None else watcher_conf_override['use_sockets']
		self.copy_env=copy_env if watcher_conf_override['copy_env'] is None else watcher_conf_override['copy_env']
		self.copy_path=copy_path if watcher_conf_override['copy_path'] is None else watcher_conf_override['copy_path']
		self.max_age=max_age if watcher_conf_override['max_age'] is None else watcher_conf_override['max_age']
		self.max_age_variance=max_age_variance if watcher_conf_override['max_age_variance'] is None else watcher_conf_override['max_age_variance']
		self.hooks=hooks if watcher_conf_override['hooks'] is None else watcher_conf_override['hooks']
		self.respawn=respawn if watcher_conf_override['respawn'] is None else watcher_conf_override['respawn']
		self.autostart=autostart if watcher_conf_override['autostart'] is None else watcher_conf_override['autostart']
		self.on_demand=on_demand if watcher_conf_override['on_demand'] is None else watcher_conf_override['on_demand']
		self.virtualenv=virtualenv if watcher_conf_override['virtualenv'] is None else watcher_conf_override['virtualenv']
		self.stdin_socket=stdin_socket if watcher_conf_override['stdin_socket'] is None else watcher_conf_override['stdin_socket']
		self.close_child_stdin=close_child_stdin if watcher_conf_override['close_child_stdin'] is None else watcher_conf_override['close_child_stdin']
		self.close_child_stdout=close_child_stdout if watcher_conf_override['close_child_stdout'] is None else watcher_conf_override['close_child_stdout']
		self.close_child_stderr=close_child_stderr if watcher_conf_override['close_child_stderr'] is None else watcher_conf_override['close_child_stderr']
		self.virtualenv_py_ver=virtualenv_py_ver if watcher_conf_override['virtualenv_py_ver'] is None else watcher_conf_override['virtualenv_py_ver']
		self.use_papa=use_papa if watcher_conf_override['use_papa'] is None else watcher_conf_override['use_papa']
		# endregion
		
		TheWatcher.watchers.append(self.to_dict())
	
	def to_dict(self):
		return {k:v for k,v in self.__dict__.items() if v is not None}

celery_default_log_level='info'

class CeleryMan(object):
	
	def __init__(self,celery_path,app_name,add_cmd=''):
		self.celery_path=celery_path
		self.app_name=app_name
		self.add_cmd=add_cmd
	
	def create_celery_worker(self,worker_name,Ofair=True,queues=None,queue_as_worker_name=True,
	                         concurrency=None,autoscale=None,log_level=None,add_cmd='',**watcher_kwargs):
		"""
		This method creates celery worker.
		
		:param add_cmd: Additional cmd to add manually to the end of constructed cmd.
		:param worker_name:
					Name of the worker. It will be provided as "-n <worker_name>" argument to celery worker.
		:param Ofair:
					see celery docs.
		:param queues:
					You can specify: 1) single queue (str), 2) list of strings as queue names.
		:param queue_as_worker_name:
					If "queues" is not provided, "worker_name" is used as queue name (by_default).
					If you don't want to specify any queue at all, then pass False.
		:param concurrency:
					Number of concurrent processses. see celery docs.
		:param autoscale:
					Tuple in the form (max,min) processes as specified in celery docs.
		:param log_level:
					Celery log level. By default = "info".
		:param watcher_kwargs:
					Other parameters are passed to TheWatcher.
		:return: Nothing.
		"""
		log_level=log_level or celery_default_log_level
		watcher_name=worker_name+'_watcher'
		parts=list()
		parts.append(f"{self.celery_path} worker -A {self.app_name} -l {log_level}")
		parts.append(f"-n {worker_name}")
		if Ofair: parts.append('-Ofair')
		if queues:
			if type(queues) is str:
				parts.append(f'-Q {queues}')
			else:
				parts.append(f'-Q {",".join(queues)}')
		elif queue_as_worker_name:
			parts.append(f'-Q {worker_name}')
		
		if autoscale:
			parts.append(f"--autoscale={','.join(map(str,autoscale))}")
		elif concurrency:
			parts.append(f'-c {concurrency}')
		
		if self.add_cmd: parts.append(self.add_cmd)
		if add_cmd: parts.append(add_cmd)
		
		TheWatcher(watcher_name," ".join(parts),**watcher_kwargs)
		pass

def run_watchers():
	arbiter=circus.get_arbiter(TheWatcher.watchers)
	try:
		arbiter.start()
	finally:
		arbiter.stop()
	return

def main():
	"""
	TheWatcher is the main thing that you need for running your commands continuously.
	
		 TheWatcher(name='some_watcher_name',
		            cmd='Command that needs to be run continuously',
		            working_dir='specify_your_working_dir',
		            numprocesses=1,
		            respawn=True)
		 
		 TheWatcher(name='another_watcher_name',
		            cmd='Another command that needs to be run continuously',
		            working_dir='some_other_working_dir',
		            numprocesses=1,
		            respawn=False)
		 
		 run_watchers()
		 
		 That's it. When you call "run_watchers", all declared watchers will be launched in parallel.
		 
		 
	 
	If you need to run celery workers, you can use CeleryMan!:
	
		For example:
			
			cman=CeleryMan(celery_path=r'/usr/local/bin/celery',app_name='celery_worker')
			
			cman.create_celery_worker('get_image_date',autoscale=(18,3))
			cman.create_celery_worker('process_main_request',autoscale=(36,3))

			cman.create_celery_worker('download_products',concurrency=1)
			cman.create_celery_worker('atmospheric_correction_of_single_1c',concurrency=2)
	"""
	TheWatcher('testik','ls -lah',respawn=True)  # plohoi primer. Eta commanda ne ranitsya continuously))).
	run_watchers()
	pass

if __name__=='__main__':
	main()



import os
import logging
import uuid

import matplotlib.pyplot as plt
import numpy as np
from osgeo import gdal,gdal_array,ogr,osr

# from . import log_support as logsup
from . import global_support as gsup
from lgblkb_tools import TheLogger,log_wrapper

simple_logger=TheLogger(__name__)
gdal.UseExceptions()

def ds_to_array(ds,band_index=1):
	if isinstance(ds,str): ds=gdal.Open(ds)
	return ds.GetRasterBand(band_index).ReadAsArray(0,0,ds.RasterXSize,ds.RasterYSize)

geom_from_ds=lambda ds:np.abs(np.array(ds.GetGeoTransform()).reshape(2,-1))
resol_from_ds=lambda ds:geom_from_ds(ds)[:,1:].sum(axis=0)
ul_from_ds=lambda ds:geom_from_ds(ds)[:,0]

get_epsg_from=lambda ds:int(
	gdal.Info(ds,format='json')['coordinateSystem']['wkt'].rsplit('"EPSG","',1)[-1].split('"')[0])

def plot_ds(ds,show=False):
	array=ds_to_array(ds)
	plt.imshow(array)
	if show: plt.show()

code_lut={v:k for k,v in gdal_array.codes.items()}

def array_to_ds(array,parent_info):
	try:
		arr_ds=gdal_array.OpenNumPyArray(array,True)
	except:
		arr_ds=gdal_array.OpenNumPyArray(array)
	if isinstance(parent_info,DataSet): parent_info=parent_info.ds
	if isinstance(parent_info,gdal.Dataset):
		transform=parent_info.GetGeoTransform()
		projection=parent_info.GetProjection()
	else:
		transform,projection=parent_info
	arr_ds.SetGeoTransform(transform)
	arr_ds.SetProjection(projection)
	return arr_ds

def get_datasource(driver_name,fpath,field_name,field_type):
	dest_datasource=ogr.GetDriverByName(driver_name).CreateDataSource(fpath)
	dest_layer=dest_datasource.CreateLayer('',srs=None)
	dest_layer.CreateField(ogr.FieldDefn(field_name,field_type))
	return dest_datasource,dest_layer

@log_wrapper()
def polygonize(ds,driver_name,fpath,field_name='DN',field_type=ogr.OFTInteger):
	dest_datasource,dest_layer=get_datasource(driver_name,fpath,field_name,field_type)
	if field_type is ogr.OFTReal: polygonizer=gdal.FPolygonize
	else: polygonizer=gdal.Polygonize
	polygonizer(ds.GetRasterBand(1),None,dest_layer,0,['8connected=8'],callback=None)
	return dest_datasource,dest_layer

def contour_generate(ds,val_ranges,driver_name,fpath,field_name,field_type):
	contour_dsrc,contour_layer=get_datasource(driver_name,fpath,field_name,field_type)
	gdal.ContourGenerate(ds.GetRasterBand(1),0,0,val_ranges,0,0,contour_layer,0,0)
	return contour_dsrc,contour_layer

#todo:Cropping dataset
class DataSet:
	
	def __init__(self,ds,array=None,name=''):
		if isinstance(ds,str):
			name=name or '_'.join(gsup.get_splitted(os.path.splitext(ds)[0])[-2:])
			ds=gdal.Open(ds)
		self.ds=ds
		geom=geom_from_ds(self.ds)
		self.resol=geom[:,1:].sum(axis=0)
		self.ul=geom[:,0]
		self.__array=array
		self.projection=self.ds.GetProjection()
		self.transform=self.ds.GetGeoTransform()
		self.name=name
	
	@log_wrapper(log_level=logging.DEBUG,skimpy=True)
	def generate_array(self):
		self.__array=ds_to_array(self.ds)
		return self
	
	@property
	def array(self):
		if self.__array is None:
			self.generate_array()
		return self.__array
	
	@property
	def epsg(self):
		return get_epsg_from(self.ds)
	
	def plot(self,show=False):
		return plot_ds(self.ds,show=show)
	
	@log_wrapper()
	def reproject(self,to_epsg):
		"""
		A sample function to reproject and resample a GDAL dataset from within
		Python. The idea here is to reproject from one system to another, as well
		as to change the pixel size. The procedure is slightly long-winded, but
		goes like this:

		1. Set up the two Spatial Reference systems.
		2. Open the original dataset, and get the geotransform
		3. Calculate bounds of new geotransform by projecting the UL corners
		4. Calculate the number of pixels with the new projection & spacing
		5. Create an in-memory raster dataset
		6. Perform the projection
		"""
		# Define the UK OSNG, see <http://spatialreference.org/ref/epsg/27700/>
		osng=osr.SpatialReference()
		osng.ImportFromEPSG(to_epsg)
		wgs84=osr.SpatialReference()
		wgs84.ImportFromEPSG(get_epsg_from(self.ds))
		tx=osr.CoordinateTransformation(wgs84,osng)
		# Up to here, all  the projection have been defined, as well as a
		# transformation from the from to the  to :)
		# We now open the dataset
		# Get the Geotransform vector
		geo_t=self.ds.GetGeoTransform()
		x_size,y_size=self.raster_sizes
		# Work out the boundaries of the new dataset in the target projection
		(ulx,uly,ulz)=tx.TransformPoint(geo_t[0],geo_t[3])
		(lrx,lry,lrz)=tx.TransformPoint(geo_t[0]+geo_t[1]*x_size,geo_t[3]+geo_t[5]*y_size)
		# See how using 27700 and WGS84 introduces a z-value!
		# Now, we create an in-memory raster
		mem_drv=gdal.GetDriverByName('MEM')
		# The size of the raster is given the new projection and pixel spacing
		# Using the values we calculated above. Also, setting it to store one band
		# and to use Float32 data type.
		#dest=mem_drv.Create('',int((lrx-ulx)/x_size),int((uly-lry)/y_size),1,gdal.GDT_Float32)
		dtype=np.typeDict[str(self.array.dtype)]
		dest=mem_drv.Create('',x_size,y_size,1,code_lut[dtype])
		#dest=mem_drv.Create('',pixel_x_size,pixel_y_size,1,gdal.GDT_Float32)
		# Calculate the new geotransform
		resol=self.resol
		new_geo=(ulx,resol[0],geo_t[2],uly,geo_t[4],-resol[1])
		# Perform the projection/resampling
		# Set the geotransform
		dest.SetGeoTransform(new_geo)
		dest.SetProjection(osng.ExportToWkt())
		gdal.ReprojectImage(self.ds,dest,wgs84.ExportToWkt(),osng.ExportToWkt(),gdal.GRA_Bilinear)
		#gdal.ReprojectImage(self.ds,dest,wgs84.ExportToWkt(),osng.ExportToWkt())
		return DataSet(dest)
	
	@log_wrapper(log_level=logging.DEBUG,skimpy=True)
	def do_warp(self,cutline_json=None,destroy_after=True,out_epsg=4326,**kwargs):
		temp_filepath=f"/vsimem/{uuid.uuid4()}"
		
		gdal.Warp(temp_filepath,self.ds,
		          cropToCutline=not cutline_json is None,srcSRS=f'EPSG:{self.epsg}',dstSRS=f'EPSG:{out_epsg}',
		          cutlineDSName=cutline_json,**dict(dict(),**kwargs))
		ds=DataSet(temp_filepath)
		# cropped_ds=gdal.Warp('',self.ds,format='VRT',cropToCutline=cropToCutline,cutlineDSName=cutline_json,**kwargs)
		if destroy_after: self.ds=None
		return ds
	
	@property
	def geo_info(self):
		return self.transform,self.projection
	
	@property
	def raster_sizes(self):
		return self.ds.RasterXSize,self.ds.RasterYSize
	
	@classmethod
	def from_array(cls,array,geo_info):
		if isinstance(geo_info,str): geo_info=DataSet(geo_info)
		return DataSet(array_to_ds(array,geo_info),array)
	
	@log_wrapper()
	def polygonize(self,driver_name,fpath,destroy_after=True,**kwargs):
		datasource,ds_layer=polygonize(self.ds,driver_name,fpath,**kwargs)
		if not 'mem' in driver_name.lower():
			ds_layer=None
			datasource=None
		if destroy_after: self.ds=None
		return datasource,ds_layer
	
	@log_wrapper()
	def generate_contours(self,val_ranges,driver_name,fpath,field_name='DN',
	                      field_type=ogr.OFTReal,destroy_after=True):
		datasource,ds_layer=contour_generate(ds=self.ds,val_ranges=val_ranges,
		                                     driver_name=driver_name,fpath=fpath,
		                                     field_name=field_name,field_type=field_type)
		if not 'mem' in driver_name.lower():
			ds_layer=None
			datasource=None
		if destroy_after: self.ds=None
		return datasource,ds_layer
	
	@log_wrapper()
	def to_geotiff(self,filepath,no_data_value=-9999):
		band=self.ds.GetRasterBand(1)
		arr=band.ReadAsArray()
		[cols,rows]=arr.shape
		driver=gdal.GetDriverByName("GTiff")
		outdata=driver.Create(filepath,rows,cols,1,gdal.GDT_Float32)
		outdata.SetGeoTransform(self.transform)  ##sets same geotransform as input
		outdata.SetProjection(self.projection)  ##sets same projection as input
		outdata.GetRasterBand(1).WriteArray(np.where(np.isnan(arr),no_data_value,arr))
		if no_data_value is not False:
			outdata.GetRasterBand(1).SetNoDataValue(no_data_value)  ##if you want these values transparent
		outdata.FlushCache()  ##saves to disk!!
		outdata=None
		band=None
		ds=None
	
	def scale_array(self,scaler):
		scaled=scaler.fit_transform(self.array.reshape(-1,1))
		scaled_band=scaled.reshape(self.array.shape)
		out=self.from_array(scaled_band,self.geo_info)
		self.ds=None
		return out

# class RGB_DataSet(object):
#
# 	def __init__(self,*band_paths,**warp_opts):
# 		if warp_opts:
# 			dss=[DataSet(path).do_warp(**warp_opts) for path in band_paths]
# 		else:
# 			dss=[DataSet(path) for path in band_paths]
# 		self.dss=dss

# @with_logging()
# def translate_dss(self,destroy_after=True,**kwargs):
# 	temp_filepath=f"/vsimem/{uuid.uuid4()}"
# 	translate_opts=gdal.TranslateOptions(outputType=gdal.GDT_UInt16,noData=0,rgbExpand='rgb',)
# 	translated_dss=[gdal.Translate(temp_filepath,x,options=translate_opts) for x in self.dss]
# 	# cropToCutline=not cutline_json is None,srcSRS=f'EPSG:{self.epsg}',dstSRS=f'EPSG:{out_epsg}',
# 	# cutlineDSName=cutline_json,**dict(dict(warpMemoryLimit=int(4e9)),**kwargs))
# 	ds=DataSet(temp_filepath)
# 	# cropped_ds=gdal.Warp('',self.ds,format='VRT',cropToCutline=cropToCutline,cutlineDSName=cutline_json,**kwargs)
# 	if destroy_after: self.dss=None
# 	return ds

@log_wrapper(log_level=logging.DEBUG)
def translate_tiff(input_tiff,output_tiff,translate_opts: gdal.TranslateOptions) -> str:
	ds=gdal.Open(input_tiff)
	ds=gdal.Translate(output_tiff,ds,options=translate_opts)
	ds=None
	return output_tiff

@log_wrapper(log_level=logging.DEBUG)
def rgb_to_geotiff(tiff_path,*band_paths,no_data_value=-9999,dtype=gdal.GDT_Float32,**warp_opts):
	# logsup.logger.info('no_data_value: %s',no_data_value)
	band_dss=[DataSet(path).do_warp(**warp_opts).generate_array() for path in band_paths]
	band_arrays=[band_ds.array for band_ds in band_dss]
	# for path in band_paths:
	# 	step1=DataSet(path).do_warp(out_epsg=3857,srcNodata=-2000)  #
	# 	step3=step1.scale_array(RobustScaler()).scale_array(MinMaxScaler((0,255))).array
	# 	band_arrays.append(step3)
	# band_arrays=[_get_image_prepared_array(path,warp_opts=dict(out_epsg=3857),scaler=MinMaxScaler((0,255))) for path in band_paths]
	# band_arrays=gsup.ParallelTasker(_get_image_prepared_array,scaler=MinMaxScaler((0,255))).set_run_params(ds=band_paths).run(sleep_time=1e-1)
	# band_arrays=gsup.ParallelTasker(ds_to_array).set_run_params(ds=band_paths).run(sleep_time=1e-1)
	[cols,rows]=band_arrays[0].shape
	driver=gdal.GetDriverByName("GTiff")
	if len(band_paths)==3:
		options=['PHOTOMETRIC=RGB','PROFILE=GeoTIFF',]
	else:
		options=None
	# elif len(band_paths)==4:
	# 	options=['PHOTOMETRIC=RGBA','PROFILE=GeoTIFF',]
	# else:
	# 	raise NotImplementedError(f'Invalid number of input files - {len(band_paths)}.',dict(count=len(band_paths)))
	
	outdata=driver.Create(tiff_path,rows,cols,len(band_paths),dtype,options=options)
	outdata.SetGeoTransform(band_dss[0].transform)  ##sets same geotransform as input
	outdata.SetProjection(band_dss[0].projection)  ##sets same projection as input
	for i,(array,band_interpretation) in\
			enumerate(zip(band_arrays,[gdal.GCI_RedBand,gdal.GCI_GreenBand,gdal.GCI_BlueBand])):
		raster_band=outdata.GetRasterBand(i+1)
		raster_band.SetColorInterpretation(band_interpretation)
		raster_band.WriteArray(np.where(array==0,no_data_value,array))
		if no_data_value is not False:
			raster_band.SetNoDataValue(no_data_value)  ##if you want these values transparent
	
	outdata.FlushCache()  ##saves to disk!!
	outdata=None
	band=None
	ds=None
	pass

def get_mean_std_scale_params(stack: np.array,std_num=2):
	mean=np.nanmean(stack)
	std=np.nanstd(stack)
	return [mean-std_num*std,mean+std_num*std]

@log_wrapper()
def main():
	# print(get_group_cloudiness())
	# cad_info: Cadastres_Info=db_opers.SessionHelper(db_opers.Cadastres_Info).get_with_id(1380560,as_saved=True)
	# band_paths=list()
	# band_paths.append(
	# 	r'/home/lgblkb/PycharmProjects/Egistic/Downloaded_images/sentinel2/unzipped_scenes/S2B_MSIL2A_20180518T062629_N0206_R077_T42UUB_20180518T101322.SAFE/GRANULE/L2A_T42UUB_A006252_20180518T063426/IMG_DATA/R10m/T42UUB_20180518T062629_B04_10m.jp2')
	# band_paths.append(
	# 	r'/home/lgblkb/PycharmProjects/Egistic/Downloaded_images/sentinel2/unzipped_scenes/S2B_MSIL2A_20180518T062629_N0206_R077_T42UUB_20180518T101322.SAFE/GRANULE/L2A_T42UUB_A006252_20180518T063426/IMG_DATA/R10m/T42UUB_20180518T062629_B03_10m.jp2')
	# band_paths.append(
	# 	r'/home/lgblkb/PycharmProjects/Egistic/Downloaded_images/sentinel2/unzipped_scenes/S2B_MSIL2A_20180518T062629_N0206_R077_T42UUB_20180518T101322.SAFE/GRANULE/L2A_T42UUB_A006252_20180518T063426/IMG_DATA/R10m/T42UUB_20180518T062629_B02_10m.jp2')
	# no_data_value=0
	# p1=gsup.results_folder.create('test1').get_filepath('ndvi',ext='.tiff',include_datetime=True)
	# p2=gsup.results_folder.create('test1').get_filepath('ndvi_merged',ext='.tiff',include_datetime=True)
	# rgb_to_geotiff(p1,cutline_json=geojson.Feature(geometry=db_opers.to_shape(cad_info.geom)),*band_paths,no_data_value=no_data_value)
	# gsup.run_shell(gsup.gdal_merge_py,'-o',p2,'-n',str(no_data_value),'-a_nodata','nan',p1,with_popen=True)
	
	# np.random.seed(1)
	
	# val_ranges=[-1,0,0.033,0.066,0.1,0.133,0.166,0.2,0.25,0.3,0.35,0.4,0.45,0.5,0.6,0.7,0.8,0.9,1]
	# simple_logger.debug('len(val_ranges): %s',len(val_ranges))
	
	# zero_catch_window=1e-9
	# val_ranges=[-1.01,-zero_catch_window,zero_catch_window,*np.linspace(zero_catch_window,1,12)]
	# mean_vals,digitized=categorize(ndvi_ds.array,val_ranges)
	# digitized*=digitized!=2
	# mean_vals.pop(2)
	# print(mean_vals)
	# print(len(mean_vals))
	# #digitized=remove_lonely_cats(digitized)
	# field_name='Category'
	# #DataSet.from_array(digitized,ndvi_ds).polygonize('geojson','arkan5_original_12.geojson',field_name=field_name)
	# contour_dsrc,contour_layer=ndvi_ds.generate_contours(val_ranges,'Memory','')
	# feats=process_vector_layer(contour_layer,100)
	# contour_dsrc=contour_layer=None
	# save_json_features(feats,'reduced_13',cad_num=cad_num)
	return

# dat_src,dlayer=DataSet.from_array(digitized,ndvi_ds).polygonize('Memory','',
#                                                                 field_name=field_name)
# poly_collection=cleanup_polygons(dlayer,field_name,area_threshold=110,thresh_len=8)
# if 0 in poly_collection: poly_collection.pop(0)
# dat_src=dlayer=None
# #print(list(poly_collection.keys()))
# features=list()
# for cat_id,geoms in poly_collection.items():
# 	simple_logger.debug('cat_id: %s',cat_id)
# 	feat=geojson.Feature(id=cat_id,geometry=shg.MultiPolygon(geoms),
# 	                     properties=dict(DN=float(mean_vals[cat_id])))
# 	features.append(feat)
# feat_col=geojson.FeatureCollection(features,cad_num=spk.cad_land.name)
# geojson.dump(feat_col,open('arkan5_simplified_12.geojson','w'))

if __name__=='__main__':
	main()



import glob
import os
import shutil
import zipfile
from datetime import datetime
from typing import Callable

import time
from box import Box
from checksumdir import dirhash

from lgblkb_tools.locations import get_parent_dir,get_splitted,InfoDict,create_path,_make_zipfile,ZipError,get_existing_path,get_name,CopyError
from .common_utils.logging_utils import TheLogger

simple_logger=TheLogger(__name__)

class Folder(object):
	
	def __init__(self,path='',pseudo=False,parent=None,propagate_type=False,assert_exists=False):
		if isinstance(path,Folder): path=path.path
		assert isinstance(pseudo,bool)
		if not pseudo:
			if not path: path=os.getcwd()
			elif not os.path.exists(path):
				if assert_exists: raise AssertionError(f'The path {path} should exist.',dict(path=path))
				assert os.path.splitext(os.path.split(path)[-1])[-1]=='',f"Please, use folder path as input. Provided path is: \n{path}"
				os.makedirs(path)
			elif not os.path.isdir(path):
				path=get_parent_dir(path,1)
		self.path=path
		self.pseudo=pseudo
		self.__parent=parent
		self.__propagate_type=propagate_type
	
	@classmethod
	def create_for_filepath(cls,filepath):
		os.makedirs(get_parent_dir(filepath),exist_ok=True)
		return cls(filepath)
	
	@property
	def name(self):
		return os.path.split(self.path)[-1]
	
	def get_filepath(self,*name_portions,ext='',delim='_',include_depth=None,datetime_loc_index=None,iterated=False,**name_kwargs):
		parts=list()
		if include_depth not in [0,None]:
			if type(include_depth) is Callable:
				parent_parts=include_depth(get_splitted(self.path))
			elif type(include_depth) is int:
				parent_parts=get_splitted(self.path)[-include_depth:]
			elif type(include_depth) in [tuple,list]:
				parent_parts=[get_splitted(self.path)[-x] for x in include_depth]
			else:
				raise NotImplementedError(f'include_depth={include_depth}')
			parts.extend(parent_parts)
		part_portions=delim.join([str(x) for x in name_portions])
		if part_portions: parts.append(part_portions)
		part_kwargs=InfoDict(name_kwargs).get_portions()
		if part_kwargs: parts.extend(part_kwargs)
		if datetime_loc_index is not None: parts.insert(datetime_loc_index,datetime.now().strftime("%Y%m%d-%H:%M:%S"))
		assert parts,'Nothing is provided to create filepath.'
		output_path=os.path.join(self.path,delim.join(parts).replace(' ',delim)+ext)
		if iterated:
			return create_iterated_path(output_path)
		else:
			return output_path
	
	def _mkdir(self,child_dirpath):
		return create_path(self.path,child_dirpath)
	
	def create(self,*child_folders,**info_kwargs):
		parts=list()
		if child_folders: parts.extend(child_folders)
		if info_kwargs: parts.extend(InfoDict(info_kwargs).get_portions())
		assert parts,'Nothing is provided to create directory'
		child_dirpath='__'.join(map(str,parts))
		
		if self.pseudo: path=os.path.join(self.path,child_dirpath)
		else: path=self._mkdir(child_dirpath)
		if self.__propagate_type:
			return self.__class__(path,pseudo=self.pseudo,parent=self)
		else:
			return Folder(path,pseudo=self.pseudo,parent=self,propagate_type=True)
	
	def delete(self):
		if self.pseudo: raise OSError('Cannot delete when in "pseudo" mode.')
		shutil.rmtree(self.path)
	
	def clear(self):
		if self.pseudo: raise OSError('Cannot delete when in "pseudo" mode.')
		try:
			self.delete()
		except FileNotFoundError:
			simple_logger.warning('FileNotFoundError when trying to clear up the results folder. Passing on.')
		create_path(self.path)
		return self
	
	def children(self,*paths):
		return glob.glob(os.path.join(self.path,*(paths or '*')))
	
	@simple_logger.wrap(skimpy=True)
	def zip(self,zip_filepath='',save_path_formatter=None,forced=False):
		max_attempts=5
		if not zip_filepath: zipfile_basepath=self.parent().get_filepath(self.name)
		elif zip_filepath[-4:]=='.zip': zipfile_basepath=zip_filepath[:-4]
		else: zipfile_basepath=zip_filepath
		if save_path_formatter is not None: zipfile_basepath=save_path_formatter(zipfile_basepath)
		if not forced:
			fullpath=zipfile_basepath+'.zip'
			if os.path.exists(fullpath) and not zipfile.ZipFile(fullpath).testzip():
				return fullpath
		
		for i in range(max_attempts):
			fullpath=_make_zipfile(base_name=zipfile_basepath,root_dir=get_parent_dir(self.path),base_dir=self.name)
			#shutil.make_archive(base_name=zipfile_basepath,format='zip',root_dir=get_parent_dir(self.path),base_dir=self.name)
			zip_obj=zipfile.ZipFile(fullpath)
			if not zip_obj.testzip():
				return fullpath
			else:
				os.remove(fullpath)
		raise ZipError(f'Could not zip {self.path} to {zipfile_basepath}.zip.')
	
	@simple_logger.wrap(skimpy=True)
	def zip_to(self,dest_folder,zipname='',save_path_formatter=None,forced=False):
		return self.zip(Folder(dest_folder).get_filepath(zipname or self.name),save_path_formatter=save_path_formatter,forced=forced)
	
	@simple_logger.wrap(skimpy=True)
	def unzip(self,zip_filepath,create_subdir=True):
		# if save_path_formatter is None: save_path_formatter=lambda x:x
		zip_path=[zip_filepath,
		          zip_filepath+'.zip',
		          self.get_filepath(zip_filepath),
		          self.get_filepath(zip_filepath,ext='.zip')]
		zip_path=get_existing_path(zip_path)
		assert not os.path.isdir(zip_path),f'The folder "{zip_filepath}" cannot be unzipped.'
		if create_subdir:
			zipfilename=get_name(zip_path)
			# self.get_filepath(zipfilename)
			shutil.unpack_archive(zip_path,self.create(zipfilename).path,'zip')
		
		else:
			shutil.unpack_archive(zip_path,self.path,'zip')
		return self
	
	def glob_search(self,*patterns,recursive=True):
		return glob.glob(self.get_filepath(*patterns),recursive=recursive)
	
	def find_item(self,partial_name,ending='*'):
		filepath=self.glob_search(f'**/*{partial_name}{ending}')
		if filepath: return filepath[0]
		else: return ''
	
	def rename(self,new_name):
		new_path=os.path.join(get_parent_dir(self.path),new_name)
		os.rename(self.path,new_path)
		self.path=new_path
		return self
	
	def move(self,dst_path,create_subdir=True):
		if self.__propagate_type: classtype=self.__class__
		else: classtype=Folder
		if create_subdir: destination_folder=classtype(dst_path).create(self.name)
		else: destination_folder=classtype(dst_path)
		shutil.move(self.path,destination_folder.path)
		return destination_folder
	
	def parent(self,depth=1):
		if self.__propagate_type: classtype=self.__class__
		else: classtype=Folder
		return classtype(get_parent_dir(self.path,depth=depth),pseudo=self.pseudo)
	
	@simple_logger.wrap(skimpy=True)
	def copy_to(self,dst_path,create_subdir=True,forced=False):
		# if self.__propagate_type: classtype=self.__class__
		# else: classtype=partial(Folder,pseudo=True)
		if create_subdir: destination_folder=self.__class__(self.__class__(dst_path).create(self.name),
		                                                    pseudo=self.pseudo,propagate_type=self.__propagate_type)
		else: destination_folder=self.__class__(dst_path,pseudo=self.pseudo,propagate_type=self.__propagate_type)
		
		simple_logger.debug(f'Copying {self.path} to {destination_folder.path}...')
		source_hashsum=dirhash(self.path)
		if not forced and os.path.exists(destination_folder.path):
			destination_hashsum=dirhash(destination_folder.path)
			if source_hashsum==destination_hashsum: return destination_folder
		max_tries=3
		for i in range(3):
			#if os.path.exists(destination_folder.path):
			shutil.rmtree(destination_folder.path,ignore_errors=True)
			shutil.copytree(self.path,destination_folder.path)
			destination_hashsum=dirhash(destination_folder.path)
			if source_hashsum==destination_hashsum: return destination_folder
			else:
				simple_logger.warning('source_hashsum',source_hashsum)
				simple_logger.warning('destination_hashsum',destination_hashsum)
				simple_logger.warning(
					f'Warning!!!!!!!!!. Copied folder has different hashsum than its source. Try {i}/{max_tries}. Waiting for 10 sec.')
				time.sleep(10)
		raise CopyError(f'{self} could not be copied to {destination_folder}.',dict(source_folder=self.path,
		                                                                            destination_folder=destination_folder.path))
	
	def get_size_in_gb(self):
		total_size=0
		for dirpath,dirnames,filenames in os.walk(self.path):
			for f in filenames:
				fp=os.path.join(dirpath,f)
				total_size+=os.path.getsize(fp)
		return total_size*1e-9
	
	@property
	def exists(self):
		return os.path.exists(self.path)
	
	def create_iterated(self,start=None,delim='_',empty_ok=False):
		if empty_ok: validator_func=lambda existing_path:not self.__class__(existing_path).children()
		else: validator_func=lambda existing_path:existing_path==self.path
		
		new_iter_folder_path=create_iterated_path(self.path,start,delim,validator_func=validator_func)
		last_part=new_iter_folder_path.split(delim)[-1]
		if last_part.isnumeric() and int(last_part)==(start or 0):
			new_folder=self.rename(new_iter_folder_path)
		else:
			new_folder=self.__class__(new_iter_folder_path)
			if not new_folder==self and not self.children(): self.delete()
		
		# base_folder=self.rename(delim.join([self.name,f'{start}']))
		# new_path=create_iterated_path(base_folder.path,delim)
		# while base_folder.children():
		#
		# 	existing_path=base_folder.path
		# 	old_counter=base_folder.path.split('_')[-1]
		# 	new_counter=int(old_counter)+1
		# 	new_path=existing_path[:-len(old_counter)]+str(new_counter)
		# 	base_folder=self.__class__(new_path)
		return new_folder
	
	def __getitem__(self,item):
		if os.path.splitext(item)[-1]:
			return self.get_filepath(item)
		else:
			return self.create(item)
	
	def __eq__(self,other):
		if isinstance(other,str):
			return self.path==other
		else:
			return self.path==self.__class__(other).path
	
	def __setitem__(self,key,value):
		filename,ext=os.path.splitext(key)
		if isinstance(value,str): string_obj=value
		elif type(value) in [list,tuple]:
			string_obj='\n'.join(map(str,value))
		elif isinstance(value,dict):
			ext=ext or '.yaml'
			if ext in ['.yaml','.yml']:
				Box(value).to_yaml(filename=self.get_filepath(filename,ext=ext))
				return
			elif ext=='.json':
				Box(value).to_json(filename=self.get_filepath(filename,ext=ext))
				return
			else:
				string_obj='\n'.join(map(lambda kv:f"{kv[0]}: {kv[1]}",value.items()))
		
		else: string_obj=str(value)
		# def writer(filehandler):
		# 	for k,v in value.items():
		# 		filehandler.writelines([f"{k}: {v}"])
		
		with open(self.get_filepath(filename,ext=ext or '.txt'),'w') as fh:
			fh.write(str(string_obj))
	
	def __repr__(self):
		return f"{self.__class__.__name__}(r'{self.path}')"

def create_iterated_path(iterated_path: str,start=None,delim='_',validator_func=None) -> str:
	if validator_func is None: validator_func=lambda x:False
	base_path,item_name_with_ext=os.path.split(iterated_path)
	item_name,ext=os.path.splitext(item_name_with_ext)
	# parts: List[str]=item_name.split(delim)
	# the_parts: list=parts[:]
	# for i_part,part in enumerate(parts[::-1]):
	# 	if part.isnumeric():
	# 		curr_iter=int(part) if start is None else start#max(start,int(part))
	# 		while True:
	# 			the_parts[len(parts)-i_part-1]=curr_iter
	# 			new_item_name=delim.join(map(str,the_parts))
	# 			new_itempath=os.path.join(base_path,new_item_name+ext)
	# 			if not os.path.exists(new_itempath) or validator_func(new_itempath): return new_itempath
	# 			curr_iter+=1
	
	curr_iter=0 if start is None else start
	while True:
		new_item_name=delim.join([item_name,f'{curr_iter}'])
		new_itempath=os.path.join(base_path,new_item_name+ext)
		if not os.path.exists(new_itempath) or validator_func(new_itempath): return new_itempath
		curr_iter+=1
